import pytest

from ..metadata.definitions import load_icat_fields
from .fixtures.icat import *  # noqa F401


@pytest.fixture
def icat_namespace():
    metadict = dict()

    def getter(key):
        return metadict[key]

    def setter(key, value):
        metadict[key] = value

    icat_fields = load_icat_fields()
    metadata = icat_fields.namespace(getter=getter, setter=setter)

    return icat_fields, metadata, metadict



def strip_xmlns(obj):
    if isinstance(obj, dict):
        return {k: strip_xmlns(v) for k, v in obj.items() if k != "@xmlns"}
    elif isinstance(obj, list):
        return [strip_xmlns(v) for v in obj]
    else:
        return obj



import json
from typing import List
from typing import Optional

from . import defaults
from .deprecation_utils import warn_deprecated_module
from .messaging import IcatMessagingClient
from .types import ArchiveStatusLevel
from .types import ArchiveStatusLevel as StatusLevel  # noqa F401 deprecated
from .types import ArchiveStatusType
from .types import ArchiveStatusType as StatusType  # noqa F401 deprecated

warn_deprecated_module()


class IcatArchiveStatusClient:
    """Client for storing archive and restoration status in ICAT."""

    def __init__(
        self,
        queue_urls: Optional[List[str]] = None,
        queue_name: Optional[str] = None,
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        if queue_name is None:
            queue_name = defaults.ARCHIVE_QUEUE
        if queue_urls is None:
            queue_urls = defaults.ARCHIVE_BROKERS
        self._client = IcatMessagingClient(
            queue_urls, queue_name, monitor_port=monitor_port, timeout=timeout
        )

    def disconnect(self):
        self._client.disconnect()

    def send_archive_status(
        self,
        dataset_id: int,
        type: ArchiveStatusType,
        level: ArchiveStatusLevel,
        message: str,
    ):
        assert dataset_id, "ICAT requires the datasetId"
        assert type, "ICAT requires the type"
        assert level, "ICAT requires the level"
        root = {
            "datasetId": dataset_id,
            "type": type.value,
            "level": level.value,
            "message": message,
        }
        data = json.dumps(root).encode("utf-8")
        self._client.send(data)

    def check_health(self):
        """Raises an exception when not healthy"""
        self._client.check_health()



from pathlib import Path

from ..apps import store_processed
from ..apps import store_raw
from .utils.message import assert_dataset_message


def test_store_raw(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("dataset"),
        "--sample",
        "samplename",
        "-p",
        "field1=value1",
        "-p",
        "field2=[1, 2, 3]",
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_raw.main(argv)

    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": _dummy_path("dataset"),
            "name": "datasetname",
            "parameter": [
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
                {"name": "Sample_name", "value": "samplename"},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def test_store_raw_metadatafile(icat_metadata_client, icat_backend, tmpdir):
    metadatafile = str(tmpdir / "metadata.txt")

    with open(metadatafile, "w") as f:
        f.write("field1=value1\n")
        f.write("field2=[1, 2, 3]\n")

    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("dataset"),
        "--sample",
        "samplename",
        "--metadatafile",
        metadatafile,
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_raw.main(argv)

    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": _dummy_path("dataset"),
            "name": "datasetname",
            "parameter": [
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
                {"name": "Sample_name", "value": "samplename"},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def test_store_processed(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("processed"),
        "--sample",
        "samplename",
        "-p",
        "field1=value1",
        "-p",
        "field2=[1, 2, 3]",
        "--queue",
        icat_backend["metadata_urls"][0],
        "--raw",
        _dummy_path("dataset1"),
        "--raw",
        _dummy_path("dataset2"),
    ]
    store_processed.main(argv)

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": _dummy_path("processed"),
            "name": "datasetname",
            "parameter": [
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
                {"name": "Sample_name", "value": "samplename"},
                {
                    "name": "input_datasets",
                    "value": f"{_dummy_path('dataset1')},{_dummy_path('dataset2')}",
                },
            ],
            "sample": {"name": "samplename"},
        }
    }
    message = messages.get(timeout=10)
    assert messages.empty()

    assert_dataset_message(message, expected)


def _dummy_path(dirname: str) -> str:
    return str(Path.home() / dirname)



import dataclasses
import datetime
import logging
import os
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set

from esrf_pathlib import ESRFPath

from ..client.types import Dataset as IcatDataset
from . import path_utils

logger = logging.getLogger("ICAT SYNC")


@dataclasses.dataclass(eq=True)
class Dataset:
    path: str  # Path of the directory
    proposal: str
    beamline: str
    name: Optional[str]
    raw_root_dir: str
    status_reason: List[str]
    startdate: Optional[datetime.datetime] = None  # From HDF5
    enddate: Optional[datetime.datetime] = None  # From HDF5
    metadata: Optional[dict] = (
        None  # From HDF5 (times are clipped to the session time slot)
    )
    icat_dataset: Optional[IcatDataset] = None  # From ICAT

    def __post_init__(self) -> None:
        self.path = path_utils.markdir(self.path)

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Dataset":
        """Factory method to create a Dataset instance from a dictionary."""
        data = data.copy()
        icat_dataset = data.get("icat_dataset")
        if icat_dataset is not None:
            data["icat_dataset"] = IcatDataset.from_dict(icat_dataset)
        return cls(**data)

    def is_invalid(self) -> bool:
        return self.name is None or not os.path.isdir(self.path)

    def has_metadata(self) -> bool:
        return self.metadata is not None

    def to_be_uploaded(self) -> bool:
        return (
            not self.is_registered()
            and self.metadata_file
            and os.path.exists(self.metadata_file)
        )

    def is_registered(self) -> bool:
        return self.icat_dataset is not None

    def is_non_empty_dir(self) -> bool:
        if not os.path.isdir(self.path):
            return False
        return bool(os.listdir(self.path))

    def is_registered_without_files(self) -> bool:
        return (
            self.is_registered()
            and self.icat_dataset.dataset_metadata.file_count == 0
            and self.is_non_empty_dir()
        )

    def get_status(self) -> str:
        # Decreasing order of registration:
        if self.is_registered_without_files():
            return "registered_without_files"
        if self.is_registered():
            return "registered"
        if self.to_be_uploaded():
            return "not_uploaded"
        # Not registered of invalid
        if self.is_invalid() or not self.has_metadata():
            return "invalid"
        return "unregistered"

    @property
    def metadata_file(self) -> Optional[str]:
        """File stored by Bliss when ending the proposal without confirmation of registration"""
        path = ESRFPath(self.path)
        try:
            return str(path.raw_metadata_file)
        except AttributeError as ex:
            logger.warning("Dataset %r has no metadata file (%s)", self.path, ex)
            return None


@dataclasses.dataclass(eq=True)
class IcatInvestigation:
    id: str
    url: str
    search_url: str
    registered_datasets: List[IcatDataset]
    startdate: Optional[datetime.datetime] = None
    enddate: Optional[datetime.datetime] = None

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "IcatInvestigation":
        """Factory method to create an IcatInvestigation instance from a dictionary."""
        data = data.copy()
        data["registered_datasets"] = [
            IcatDataset.from_dict(dataset_data)
            for dataset_data in data["registered_datasets"]
        ]
        return cls(**data)

    @property
    def has_timeslot(self) -> bool:
        return self.startdate is not None

    @property
    def open_ended(self) -> bool:
        return self.enddate is None

    @property
    def started(self) -> bool:
        if not self.has_timeslot:
            return False
        now = datetime.datetime.now().astimezone()
        startdate = self.startdate - datetime.timedelta(days=2)
        return now >= startdate

    def _start_from_now(self) -> str:
        if not self.has_timeslot:
            return ""
        now = datetime.datetime.now().astimezone()
        if now >= self.startdate:
            return f"(Started {now - self.startdate} ago)"
        else:
            return f"(Will start in {self.startdate - now})"

    @property
    def ended(self) -> bool:
        if self.open_ended:
            return False
        now = datetime.datetime.now().astimezone()
        enddate = self.enddate + datetime.timedelta(days=2)
        return now >= enddate

    def _end_from_now(self) -> str:
        if self.open_ended:
            return ""
        now = datetime.datetime.now().astimezone()
        if now >= self.enddate:
            return f"(Ended {now - self.enddate} ago)"
        else:
            return f"(Will end in {self.enddate - now})"

    @property
    def ongoing(self) -> bool:
        return self.started and not self.ended

    @property
    def duration(self) -> Optional[datetime.timedelta]:
        if not self.open_ended:
            return self.enddate - self.startdate

    @property
    def timeslot(self) -> str:
        if self.open_ended:
            return f"{self.startdate} - ???"
        return f"{self.startdate} - {self.enddate} ({self.duration})"

    def pprint(self) -> None:
        logger.debug("Investigation ID: %s", self.id)
        logger.debug(" URL: %s", self.url)
        logger.debug(" Search URL: %s", self.search_url)

        if not self.has_timeslot:
            logger.debug("Invalid investigation (no start date)")
            return
        logger.debug(" Start time: %s %s", self.startdate, self._start_from_now())

        if self.open_ended:
            logger.debug(" Duration:", "open-ended (not official)")
            return
        logger.debug(" End time: %s %s", self.enddate, self._end_from_now())

        if self.ongoing:
            logger.debug(" Duration: %s (WARNING: ONGOING)", self.duration)
        else:
            logger.debug(" Duration: %s", self.duration)


@dataclasses.dataclass(eq=True)
class ExperimentalSession:
    session_dir: str
    raw_root_dir: str  # RAW_DATA directory
    raw_data_format: str  # e.g. "esrfv3"
    proposal: str
    beamline: str
    session: str
    startdate: datetime.date  # from the directory name
    search_url: str
    datasets: Dict[str, List[Dataset]]  # from browsing the root directory
    icat_investigation: Optional[IcatInvestigation] = None

    DATASET_STATUSES = (
        "unregistered",
        "not_uploaded",
        "registered",
        "invalid",
        "registered_without_files",
    )

    def __post_init__(self) -> None:
        self.datasets = {
            status: self.datasets.get(status, list())
            for status in self.DATASET_STATUSES
        }
        self.session_dir = path_utils.markdir(self.session_dir)
        self.raw_root_dir = path_utils.markdir(self.raw_root_dir)

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExperimentalSession":
        data = data.copy()

        data["datasets"] = {
            key: [Dataset.from_dict(dataset_data) for dataset_data in dataset_list]
            for key, dataset_list in data["datasets"].items()
        }

        icat_investigation = data.get("icat_investigation")
        if icat_investigation is not None:
            data["icat_investigation"] = IcatInvestigation.from_dict(icat_investigation)

        return cls(**data)

    @classmethod
    def allow_unsupervised_upload(cls, dataset_status: str) -> bool:
        if dataset_status not in cls.DATASET_STATUSES:
            raise ValueError(
                f"'{dataset_status}' is not as valid dataset status. Valid statuses are {list(cls.DATASET_STATUSES)}"
            )
        return dataset_status == "registered_without_files"

    @property
    def in_icat_investigation(self) -> Optional[bool]:
        if self.icat_investigation is None:
            return
        if not self.icat_investigation.has_timeslot:
            return False
        inside_timeslot = self.startdate >= self.icat_investigation.startdate.date()
        if not self.icat_investigation.open_ended:
            inside_timeslot &= self.startdate <= self.icat_investigation.enddate.date()
        return inside_timeslot

    def add_dataset(self, dataset: Dataset) -> None:
        self.datasets[dataset.get_status()].append(dataset)

    def iter_datasets(self) -> Generator[Dataset, None, None]:
        for lst in self.datasets.values():
            yield from lst

    def dataset_statuses(self) -> Set[str]:
        return set(status for status, lst in self.datasets.items() if lst)

    def dataset_status_counts(self) -> Dict[str, int]:
        return {status: len(lst) for status, lst in self.datasets.items()}

    def tabular_data(self) -> Dict[str, str]:
        data = self._init_tabular_data()
        self._add_icat_tabular_data(data)
        return data

    def _init_tabular_data(self) -> Dict[str, str]:
        status_counts = {
            status: str(counts)
            for status, counts in self.dataset_status_counts().items()
        }
        return {
            "beamline": self.beamline,
            "proposal": self.proposal,
            "session": str(self.startdate),
            "url": "",
            "search_url": self.search_url,
            "path": self.raw_root_dir,
            **status_counts,
            "timeslot": "",
        }

    def _add_icat_tabular_data(self, data: Dict[str, str]) -> None:
        if self.icat_investigation is None:
            return
        data["url"] = self.icat_investigation.url
        data["search_url"] = self.icat_investigation.search_url

        if self.in_icat_investigation:
            data["session"] += " (NOT INSIDE TIMESLOT)"
        if not self.icat_investigation.started:
            data["timeslot"] = f"FUTURE: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.ended:
            data["timeslot"] = f"FINISHED: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.open_ended:
            data["timeslot"] = f"INDEFINITE: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.ongoing:
            data["timeslot"] = f"ONGOING: {self.icat_investigation.timeslot}"
        else:
            raise RuntimeError("Unknown session time state")

        return data

    def pprint(self) -> None:
        separator = "--------------------------------"
        logger.debug("")
        logger.debug(separator)
        self._pprint_datasets()
        logger.debug("")
        self._pprint_raw_session()
        logger.debug("")
        self._pprint_icat_investigation()
        logger.debug("")
        logger.debug(separator)
        logger.debug("")

    def _pprint_raw_session(self) -> None:
        logger.debug("Directory: %s", self.raw_root_dir)
        if self.in_icat_investigation:
            logger.debug(" Start time: %s", self.startdate)
        else:
            logger.debug(
                " Start time: %s (WARNING: not inside time slot!)", self.startdate
            )

    def _pprint_icat_investigation(self) -> None:
        if self.icat_investigation is None:
            logger.debug("No corresponding investigation")
            logger.debug(" Search URL: %s", self.search_url)
        else:
            self.icat_investigation.pprint()

    def _pprint_datasets(self) -> None:
        _print_dataset_summary("Registered datasets", self.datasets["registered"])
        _print_dataset_summary("Invalid datasets", self.datasets["invalid"])
        _print_dataset_summary("Unregistered datasets", self.datasets["unregistered"])
        _print_dataset_summary("Datasets not uploaded", self.datasets["not_uploaded"])
        _print_dataset_summary(
            "Registered datasets without files",
            self.datasets["registered_without_files"],
        )

        logger.debug("")
        if (
            self.datasets["invalid"]
            or self.datasets["not_uploaded"]
            or self.datasets["unregistered"]
            or self.datasets["registered_without_files"]
        ):
            logger.debug("Datasets (TODO):")
        elif not self.datasets["registered"]:
            logger.debug("Datasets (EMPTY):")
        else:
            logger.debug("Datasets (OK):")
        logger.debug(" %s registered", len(self.datasets["registered"]))
        logger.debug(" %s invalid", len(self.datasets["invalid"]))
        logger.debug(" %s not uploaded", len(self.datasets["not_uploaded"]))
        logger.debug(" %s unregistered", len(self.datasets["unregistered"]))
        logger.debug(
            " %s registered without files",
            len(self.datasets["registered_without_files"]),
        )


def _print_dataset_summary(title: str, datasets: List[Dataset]):
    if not datasets:
        return
    logger.debug("")
    logger.debug("%s:", title)
    for dataset in datasets:
        logger.debug("  %s", dataset.path)
        for s in dataset.status_reason:
            logger.debug("   %s", s)



import json
from typing import List
from typing import Optional

from . import defaults
from .messaging import IcatMessagingClient


class IcatUpdateMetadataClient:
    """Client for updating the metadata linked to datasets."""

    def __init__(
        self,
        queue_urls: Optional[List[str]] = None,
        queue_name: Optional[str] = None,
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        if queue_name is None:
            queue_name = defaults.UPDATE_METADATA_QUEUE
        if queue_urls is None:
            queue_urls = defaults.UPDATE_METADATA_BROKERS
        self._client = IcatMessagingClient(
            queue_urls, queue_name, monitor_port=monitor_port, timeout=timeout
        )

    def disconnect(self):
        self._client.disconnect()

    def send_update_metadata(
        self,
        proposal: str,
        beamline: str,
        dataset_paths: List[str],
        metadata_name: str,
        metadata_value: str,
    ):
        assert proposal, "ICAT requires the proposal name"
        assert beamline, "ICAT requires the beamline name"
        assert dataset_paths, "ICAT requires the dataset paths"
        assert metadata_name, "ICAT requires the metadata name"
        assert metadata_value, "ICAT requires the metadata value"
        root = {
            "proposal": proposal,
            "beamline": beamline,
            "datasetPaths": dataset_paths,
            "metadataName": metadata_name,
            "metadataValue": metadata_value,
        }
        data = json.dumps(root).encode("utf-8")
        self._client.send(data)

    def check_health(self):
        """Raises an exception when not healthy"""
        self._client.check_health()



import argparse
import datetime
import logging
import os
import sys
from glob import glob
from time import sleep
from typing import Generator
from typing import Optional

from tqdm import tqdm

from ..client.bliss import get_icat_client
from ..client.main import IcatClient
from ..utils import path_utils
from ..utils import raw_data
from ..utils import sync_types
from ..utils.log_utils import basic_config
from ..utils.sync_store import ExperimentalSessionStore

logger = logging.getLogger("ICAT SYNC")


def sync_raw(
    icat_client: IcatClient,
    beamline: Optional[str] = None,
    proposal: Optional[str] = None,
    session: Optional[str] = None,
    root_dir: Optional[str] = None,
    dry_run: bool = True,
    unsupervised: bool = False,
    raw_data_format: str = "esrfv3",
    save_dir: Optional[str] = None,
    cache_dir: Optional[str] = None,
    invalidate_cache: bool = False,
) -> None:
    with ExperimentalSessionStore(
        cache_dir=cache_dir,
        save_dir=save_dir,
        raw_data_format=raw_data_format,
        invalidate_cache=invalidate_cache,
    ) as exp_session_store:
        if beamline is None:
            beamline = "*"
        if proposal is None:
            proposal = "*"
        if session is None:
            session = "*"
        session_filter = raw_data.get_session_dir(
            proposal,
            beamline,
            session,
            root_dir=root_dir,
            raw_data_format=raw_data_format,
        )

        logger.info("Discovering sessions %s ...", session_filter)
        session_dirs = glob(path_utils.markdir(session_filter))
        logger.info("Discovering sessions finished.")

        logger.info("Parse %d session directories ...", len(session_dirs))

        if dry_run and logger.getEffectiveLevel() > logging.DEBUG:
            session_dirs = tqdm(session_dirs)

        for session_dir in session_dirs:
            (
                proposal_from_dir,
                beamline_from_dir,
                session_from_dir,
            ) = raw_data.parse_session_dir(session_dir, raw_data_format)
            if not proposal_from_dir:
                continue
            session = path_utils.basename(session_dir)
            try:
                _ = sync_session(
                    icat_client,
                    proposal_from_dir,
                    beamline_from_dir,
                    session_from_dir,
                    dry_run=dry_run,
                    unsupervised=unsupervised,
                    raw_data_format=raw_data_format,
                    exp_session_store=exp_session_store,
                )
            except Exception as e:
                raise RuntimeError(
                    f"Failed syncing beamline={beamline_from_dir}, proposal={proposal_from_dir}, session={session_from_dir}"
                ) from e
        logger.info("Session directories parsed.")


def sync_session(
    icat_client: IcatClient,
    proposal: str,
    beamline: str,
    session: str,
    root_dir: Optional[str] = None,
    dry_run: bool = True,
    unsupervised: bool = False,
    allow_open_ended: bool = True,
    raw_data_format: str = "esrfv3",
    exp_session_store: Optional[ExperimentalSessionStore] = None,
) -> Optional[sync_types.ExperimentalSession]:
    remove_from_cache_afterwards = False
    exp_session = None
    try:
        # Experimental session from cache
        if exp_session_store is not None:
            session_dir = raw_data.get_session_dir(
                proposal,
                beamline,
                session,
                root_dir=root_dir,
                raw_data_format=raw_data_format,
            )
            exp_session = exp_session_store.get_session(session_dir)

        # Experimental session from disk and ICAT
        if exp_session is None:
            exp_session = _discover_exp_session(
                icat_client,
                proposal,
                beamline,
                session,
                root_dir=root_dir,
                raw_data_format=raw_data_format,
            )

        if exp_session is None:
            # Not an experimental session
            return exp_session

        # Experimental session from ICAT
        if exp_session.icat_investigation is None:
            _discover_icat_investigation(
                icat_client,
                exp_session,
                allow_open_ended=allow_open_ended,
                raw_data_format=raw_data_format,
            )

        # Print summary of the experimental session
        exp_session.pprint()

        # Register datasets that failed to be registered
        for dataset in _iter_datasets_to_upload(
            exp_session, "not_uploaded", dry_run, unsupervised
        ):
            remove_from_cache_afterwards = True
            icat_client.store_dataset_from_file(dataset.metadata_file)
            sleep(0.1)  # do not overload ICAT

        # Register datasets that were not even attempted to be registered
        for dataset in _iter_datasets_to_upload(
            exp_session, "unregistered", dry_run, unsupervised
        ):
            remove_from_cache_afterwards = True
            icat_client.store_dataset(
                beamline=dataset.beamline,
                proposal=dataset.proposal,
                dataset=dataset.name,
                path=dataset.path,
                metadata=dataset.metadata,
            )
            sleep(0.1)  # do not overload ICAT

        # Add missing data files for registered datasets without data
        for dataset in _iter_datasets_to_upload(
            exp_session, "registered_without_files", dry_run, unsupervised
        ):
            remove_from_cache_afterwards = True
            icat_client.add_files(dataset.icat_dataset.icat_dataset_id)
            sleep(0.1)  # do not overload ICAT

    finally:
        if exp_session_store and exp_session:
            if remove_from_cache_afterwards:
                exp_session_store.remove_session(exp_session)
            else:
                exp_session_store.add_session(exp_session)

    return exp_session


def _iter_datasets_to_upload(
    exp_session: sync_types.ExperimentalSession,
    dataset_status: str,
    dry_run: bool,
    unsupervised: bool,
    verbose: bool = False,
) -> Generator[sync_types.Dataset, None, None]:
    datasets = exp_session.datasets[dataset_status]
    if not datasets:
        return

    action = f'Upload "{dataset_status}" dataset'

    if dry_run:
        do_yield = False
        do_print = verbose
    elif unsupervised:
        if exp_session.allow_unsupervised_upload(dataset_status):
            do_yield = True
            do_print = True
        else:
            do_yield = False
            do_print = verbose
    else:
        if _ask_execute_confirmation(exp_session, action + "s"):
            do_yield = True
            do_print = True
        else:
            do_yield = False
            do_print = verbose

    if not do_yield and not do_print:
        return

    for dataset in datasets:
        if do_print:
            _print_dataset_details(dataset, action)
        if do_yield:
            yield dataset


def _ask_execute_confirmation(
    exp_session: sync_types.ExperimentalSession, action: str
) -> bool:
    logger.info("Search URL: %s", exp_session.icat_investigation.search_url)
    result = input(f"{action}? (y/[n])")
    return result.lower() in ("y", "yes")


def _print_dataset_details(dataset: sync_types.Dataset, title: str):
    logger.debug("-> %s: %s", title, dataset.path)
    logger.debug("    Name: %s", dataset.name)
    if dataset.metadata:
        logger.debug("    Sample: %s", dataset.metadata["Sample_name"])
        dataset_startdate = dataset.startdate
        dataset_enddate = dataset.enddate
        logger.debug("    Start time: %s", dataset_startdate)
        logger.debug("    End time: %s", dataset_enddate)
        logger.debug("    Duration: %s", dataset_enddate - dataset_startdate)


def _discover_exp_session(
    icat_client: IcatClient,
    proposal: str,
    beamline: str,
    session: str,
    root_dir: Optional[str] = None,
    raw_data_format: str = "esrfv3",
) -> Optional[sync_types.ExperimentalSession]:
    """Discovery the experimental session on disk and in ICAT"""
    try:
        session_startdate_fromdir = datetime.datetime.strptime(session, "%Y%m%d").date()
    except ValueError:
        # Not a session directory
        return

    session_dir = raw_data.get_session_dir(
        proposal, beamline, session, root_dir=root_dir, raw_data_format=raw_data_format
    )

    raw_root_dir = raw_data.get_raw_data_dir(
        session_dir, raw_data_format=raw_data_format
    )

    return sync_types.ExperimentalSession(
        session_dir=session_dir,
        raw_root_dir=raw_root_dir,
        raw_data_format=raw_data_format,
        proposal=proposal,
        beamline=beamline,
        session=session,
        startdate=session_startdate_fromdir,
        search_url=_data_portal_search_url(proposal, beamline),
        datasets=dict(),
    )


def _discover_icat_investigation(
    icat_client: IcatClient,
    exp_session: sync_types.ExperimentalSession,
    allow_open_ended: bool = True,
    raw_data_format: str = "esrfv3",
) -> None:
    # Get the ICAT investigation related to the raw data directory
    investigation = icat_client.investigation_info(
        exp_session.beamline,
        exp_session.proposal,
        date=exp_session.startdate,
        allow_open_ended=allow_open_ended,
    )
    if not investigation:
        return
    if not investigation.get("startDate"):
        return

    icat_investigation = sync_types.IcatInvestigation(
        id=investigation["id"],
        url=investigation["data portal"],
        search_url=_data_portal_search_url(
            investigation["proposal"], investigation["beamline"]
        ),
        registered_datasets=list(),
    )

    icat_investigation.startdate = _as_datetime(investigation["startDate"])
    if investigation.get("endDate"):
        icat_investigation.enddate = _as_datetime(investigation["endDate"])

    # Add dataset information
    exp_session.icat_investigation = icat_investigation
    _discover_icat_datasets(icat_client, exp_session, allow_open_ended=allow_open_ended)
    _discover_raw_datasets(exp_session, raw_data_format=raw_data_format)


def _discover_icat_datasets(
    icat_client: IcatClient,
    exp_session: sync_types.ExperimentalSession,
    allow_open_ended: bool = True,
) -> None:
    logger.info("Extracting metadata from ICAT for %s ...", exp_session.raw_root_dir)
    exp_session.icat_investigation.registered_datasets = (
        icat_client.registered_datasets(
            exp_session.beamline,
            exp_session.proposal,
            date=exp_session.startdate,
            allow_open_ended=allow_open_ended,
        )
    )
    logger.info("Extracting metadata finished.")


def _discover_raw_datasets(
    exp_session: sync_types.ExperimentalSession,
    raw_data_format: str = "esrfv3",
) -> None:
    """Browse all datasets from the raw data directory and compare with
    the datasets registered with the ICAT investigation"""

    dataset_filters = raw_data.get_dataset_filters(
        exp_session.raw_root_dir, raw_data_format=raw_data_format
    )

    if len(dataset_filters) > 1:
        logger.info(
            "Extracting metadata from HDF5 for %d filters ...", len(dataset_filters)
        )
    elif len(dataset_filters) == 1:
        logger.info("Extracting metadata from HDF5 for %s ...", dataset_filters[0])
    else:
        logger.warning("No dataset filters provided, skipping metadata extraction.")

    registered_datasets = {
        path_utils.markdir(dset.dataset_id.path): dset
        for dset in exp_session.icat_investigation.registered_datasets
    }

    dataset_dirs = []
    for pattern in dataset_filters:
        dataset_dirs.extend(glob(path_utils.markdir(pattern)))
    dataset_dirs.sort(key=_get_creation_time)

    for dataset_dir in dataset_dirs:
        name = raw_data.get_raw_dataset_name(
            dataset_dir, raw_data_format=raw_data_format
        )
        if not os.path.isdir(dataset_dir):
            status_reason = ["Dataset path is not a directory"]
        elif not name:
            status_reason = ["Dataset path has the wrong format"]
        else:
            status_reason = []

        dataset = sync_types.Dataset(
            path=dataset_dir,
            proposal=exp_session.proposal,
            beamline=exp_session.beamline,
            name=name,
            raw_root_dir=exp_session.raw_root_dir,
            status_reason=status_reason,
        )

        if not dataset.is_invalid():
            dataset.icat_dataset = registered_datasets.get(dataset.path)
            if not dataset.is_registered():
                _add_metadata_from_raw_data(
                    dataset, exp_session, raw_data_format=raw_data_format
                )

        exp_session.add_dataset(dataset)
    logger.info("Metadata from HDF5 extracted.")


def _get_creation_time(file_path: str) -> float:
    """
    Get the creation time of a file or directory.
    """
    if os.path.exists(file_path):
        return os.path.getctime(file_path)
    return 0


def _data_portal_search_url(
    proposal: str,
    beamline: str,
) -> str:
    return f"https://data.esrf.fr/beamline/{beamline}?search={proposal}"


def _add_metadata_from_raw_data(
    dataset: sync_types.Dataset,
    exp_session: sync_types.ExperimentalSession,
    raw_data_format: str = "esrfv3",
) -> None:
    try:
        dataset_metadata = raw_data.get_raw_dataset_metadata(
            dataset.path, raw_data_format=raw_data_format
        )
        err_msg = f"data format assumed to be '{raw_data_format}'"
    except Exception as e:
        dataset_metadata = dict()
        err_msg = str(e)
    minimal_set_metadata = {"Sample_name", "startDate", "endDate"}
    if not minimal_set_metadata.issubset(dataset_metadata):
        dataset.status_reason.append(f"Cannot extract dataset metadata: {err_msg}")
        return

    dataset.startdate = _as_datetime(dataset_metadata["startDate"])
    dataset.enddate = _as_datetime(dataset_metadata["endDate"])

    msg = _check_date_within_range(
        exp_session.icat_investigation.startdate,
        exp_session.icat_investigation.enddate,
        dataset.startdate,
        end=False,
    )
    if msg:
        dataset.status_reason.append(msg)

    msg = _check_date_within_range(
        exp_session.icat_investigation.startdate,
        exp_session.icat_investigation.enddate,
        dataset.enddate,
        end=True,
    )
    if msg:
        dataset.status_reason.append(msg)

    dataset_metadata["startDate"] = dataset.startdate
    dataset_metadata["endDate"] = dataset.enddate
    dataset_metadata["investigationId"] = exp_session.icat_investigation.id
    dataset.metadata = dataset_metadata


def _check_date_within_range(
    session_startdate: datetime.datetime,
    session_enddate: Optional[datetime.datetime],
    dataset_date: datetime.datetime,
    end: bool = False,
) -> Optional[str]:
    if end:
        action = "ended"
    else:
        action = "started"

    if session_enddate is not None and dataset_date >= session_enddate:
        return f"{action} {dataset_date - session_enddate} after the end of the session"
    if dataset_date <= session_startdate:
        return f"{action} {session_startdate - dataset_date} before the start of the session"

    return None


def _as_datetime(isoformat: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(isoformat).astimezone()


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Register missing raw dataset with ICAT"
    )
    parser.add_argument(
        "--beamline", type=str.lower, required=False, help="Beamline name (e.g. id00)"
    )
    parser.add_argument(
        "--proposal",
        type=str.lower,
        required=False,
        help="Proposal name (e.g. ihch123)",
    )
    parser.add_argument(
        "--session", required=False, help="Session name (e.g. 20231028)"
    )
    parser.add_argument(
        "--register",
        action="store_true",
        required=False,
        help="Register datasets with ICAT when needed (prompts for confirmation)",
    )
    parser.add_argument(
        "--auto-register",
        action="store_true",
        required=False,
        help="Register datasets with ICAT when needed without confirmation (only when it is safe to do so)",
    )
    parser.add_argument(
        "--format",
        required=False,
        choices=["esrfv1", "esrfv2", "esrfv3", "id16bspec", "mx"],
        default="esrfv3",
        help="Raw data structure",
    )
    parser.add_argument(
        "--save-dir",
        type=str,
        default=None,
        required=False,
        help="Directory to save the sync statistics for humans",
    )
    parser.add_argument(
        "--cache-dir",
        type=str,
        default=None,
        required=False,
        help="Directory to cache the sync state for machines",
    )
    parser.add_argument(
        "--no-print", action="store_true", help="Do not print the session summaries"
    )
    parser.add_argument(
        "--invalidate-cache",
        action="store_true",
        help="Invalidate session when loading the cache",
    )
    parser.add_argument(
        "--queue",
        dest="metadata_urls",
        action="append",
        help="ActiveMQ queue URLS",
        default=[],
    )
    args = parser.parse_args(argv[1:])

    basic_config(
        logger=logger,
        level=logging.INFO if args.no_print else logging.DEBUG,
        format="%(message)s",
    )

    icat_client = get_icat_client(timeout=600, metadata_urls=args.metadata_urls)

    dry_run = not args.register and not args.auto_register
    unsupervised = args.auto_register

    sync_raw(
        icat_client,
        beamline=args.beamline,
        proposal=args.proposal,
        session=args.session,
        dry_run=dry_run,
        unsupervised=unsupervised,
        raw_data_format=args.format,
        save_dir=args.save_dir,
        cache_dir=args.cache_dir,
        invalidate_cache=args.invalidate_cache,
    )

    icat_client.disconnect()


if __name__ == "__main__":
    sys.exit(main())



import base64
import datetime
import logging
import mimetypes
import socket
import warnings
from enum import Enum
from importlib.metadata import version as get_version
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from urllib.parse import urljoin

import requests

from ..utils.url import normalize_url
from . import defaults

release = get_version("pyicat_plus")

logger = logging.getLogger(__name__)


class IcatElogbookClient:
    """Client for the e-logbook part of the ICAT+ REST API.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
        **payload,
    ):
        if api_key is None:
            api_key = defaults.ELOGBOOK_TOKEN
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = f"dataacquisition/{api_key}/notification"
        self._message_url = urljoin(url, path)

        path = f"dataacquisition/{api_key}/base64"
        self._data_url = urljoin(url, path)

        self._init_payload = payload
        self._init_payload.setdefault("machine", socket.getfqdn())
        self._init_payload.setdefault("software", "pyicat-plus_v" + release)

        self.raise_error = True
        if timeout is None:
            timeout = 0.1
        self.timeout = timeout

    def _merge_payloads(self, message_payload: dict, call_payload: dict) -> dict:
        payloads = self._sorted_payloads(message_payload, call_payload)
        result = {k: v for payload in payloads for k, v in payload.items()}
        tags = self._merge_payload_tags(*payloads)
        if tags:
            result.pop("tags", None)
            result["tag"] = tags
        return result

    def _sorted_payloads(self, message_payload: dict, call_payload: dict) -> List[dict]:
        """Sorted by increasing priority"""
        return [message_payload, self._init_payload, call_payload]

    def _merge_payload_tags(self, *payloads: Iterable[dict]) -> List[dict]:
        """The payload tags can be eithers a list of strings or a list of dictionaries.
        The return value are the merged tags as a list of dictionaries.
        """
        names = set()
        tags = list()
        for payload in payloads:
            ptags = payload.get("tag", list()) + payload.get("tags", list())
            for tag in ptags:
                if isinstance(tag, str):
                    if tag in names:
                        continue
                    names.add(tag)
                    tags.append({"name": tag})
                else:
                    if tag["name"] in names:
                        continue
                    names.add(tag["name"])
                    tags.append(tag)
        return tags

    def _post_with_payload(
        self, url: str, message_payload: dict, call_payload: dict
    ) -> None:
        payload = self._merge_payloads(message_payload, call_payload)
        payload.setdefault(
            "creationDate", datetime.datetime.now().astimezone().isoformat()
        )
        try:
            response = requests.post(url, json=payload, timeout=self.timeout)
        except requests.exceptions.ReadTimeout:
            return  # we have no confirmation that the call succeeded
        except Exception as e:
            if self.raise_error:
                raise
            logger.exception(e)
            return
        if self.raise_error:
            response.raise_for_status()
        elif not response.ok:
            logger.error("%s: %s", response, response.text)

    def send_message(
        self,
        message: str,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        investigation_id: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._message_url,
            beamline=beamline,
            proposal=proposal,
            investigation_id=investigation_id,
        )
        message_payload = self._encode_message(
            message,
            message_type=message_type,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            dataset=dataset,
        )
        self._post_with_payload(url, message_payload, call_payload)

    def send_binary_data(
        self,
        data: bytes,
        mimetype: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._data_url,
            beamline=beamline,
            proposal=proposal,
            investigation_id=investigation_id,
        )
        message_payload = self._encode_binary_data(data, mimetype=mimetype)
        self._post_with_payload(url, message_payload, call_payload)

    @staticmethod
    def _compose_url(
        url: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
    ):
        query = {}
        if beamline:
            query["instrumentName"] = beamline
        if proposal:
            query["investigationName"] = proposal
        if investigation_id:
            query["investigationId"] = investigation_id
        query = "&".join([f"{k}={v}" for k, v in query.items()])
        return f"{url}?{query}"

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        dataset: Optional[str] = None,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        **payload,
    ):
        with open(filename, "r") as f:
            message = f.read()
        self.send_message(
            message,
            message_type=message_type,
            proposal=proposal,
            investigation_id=investigation_id,
            beamline=beamline,
            dataset=dataset,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            **payload,
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        **payload,
    ):
        with open(filename, "rb") as f:
            data = f.read()
        mimetype, _ = mimetypes.guess_type(filename, strict=True)
        self.send_binary_data(
            data, mimetype=mimetype, beamline=beamline, proposal=proposal, **payload
        )

    def _encode_message(
        self,
        message: str,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        dataset: Optional[str] = None,
    ) -> dict:
        message_category, message_type = _message_category_and_type(
            message_type=message_type, editable=editable, formatted=formatted
        )
        if mimetype is None:
            mimetype = "text/plain"
        try:
            format = _MessageFormatMapping[mimetype]
        except KeyError:
            raise ValueError(
                f"mime type '{mimetype}' is not supported ({list(_MessageFormatMapping)})"
            ) from None
        message = {
            "type": message_type.name,
            "category": message_category.name,
            "content": [{"format": format.name, "text": message}],
        }
        if dataset:
            message["datasetName"] = dataset
        return message

    def _encode_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
    ) -> dict:
        if not mimetype:
            # arbitrary binary data
            mimetype = "application/octet-stream"
        data_header = f"data:{mimetype};base64,"
        data_blob = base64.b64encode(data).decode("latin-1")
        return {"base64": data_header + data_blob}


_MessageCategory = Enum("MessageCategory", "debug info error commandLine comment")
_MessageType = Enum("MessageType", "annotation notification")
_MessageFormat = Enum("MessageType", "plainText html")

_MessageCategoryMapping = {
    "debug": _MessageCategory.debug,
    "info": _MessageCategory.info,
    "warning": _MessageCategory.error,
    "warn": _MessageCategory.error,
    "error": _MessageCategory.error,
    "critical": _MessageCategory.error,
    "fatal": _MessageCategory.error,
    "command": _MessageCategory.commandLine,
    "comment": _MessageCategory.comment,
}

_MessageFormatMapping = {
    "text/plain": _MessageFormat.plainText,
    "text/html": _MessageFormat.html,
}


def _message_category_and_type(
    message_type: Optional[str] = None,
    editable: Optional[bool] = None,
    formatted: Optional[bool] = None,
) -> Tuple[_MessageCategory, _MessageType]:
    """Derive the ICAT message category from the API message type.

    The ICAT message types are:

    - "annotation" (default for comments): editable and unformatted message
    - "notification" (default for non-comments): formatted and not editable

    Only comments can be editable and a message cannot be editable and
    formatted at the same time.
    """
    if message_type is None:
        message_type = "comment"
    try:
        category = _MessageCategoryMapping[message_type.lower()]
    except KeyError:
        raise ValueError(
            f"'{message_type}' is not a valid e-logbook message type"
        ) from None

    if category != _MessageCategory.comment:
        # Non-comments cannot be editable
        if editable:
            warnings.warn(
                f"message type '{message_type}' cannot be editable", UserWarning
            )
            editable = None

        # Non-comments cannot be unformatted
        if formatted is not None and not formatted:
            warnings.warn(
                f"message type '{message_type}' cannot be unformatted", UserWarning
            )
            formatted = None

    # A message cannot be editable and formatted at the same time or uneditable and unformatted
    if formatted == editable and formatted is not None:
        if formatted:
            warnings.warn(
                f"message type '{message_type}' cannot be editable and formatted at the same time",
                UserWarning,
            )
        else:
            warnings.warn(
                f"message type '{message_type}' cannot be uneditable and unformatted at the same time",
                UserWarning,
            )
        formatted = None
        editable = None

    # Editability is specified
    if editable is not None:
        if editable:
            return category, _MessageType.annotation
        return category, _MessageType.notification

    # Formatability is specified
    if formatted is not None:
        if formatted:
            return category, _MessageType.notification
        return category, _MessageType.annotation

    # By default comments are annotations and the rest are notifications
    if category == _MessageCategory.comment:
        return category, _MessageType.annotation
    return category, _MessageType.notification



import datetime
from typing import List
from typing import Optional
from urllib.parse import urljoin

import requests

from ..utils.url import normalize_url
from .deprecation_utils import deprecated_argument
from .types import SampleMetadata


class IcatPlusRestrictedClient:
    """Client for the restricted part of the ICAT+ REST API.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self, url: str, password: Optional[str] = None, session_id: Optional[str] = None
    ):
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = "catalogue/{session_id}/investigation"
        self._investigation_url = urljoin(url, path)

        path = "catalogue/{session_id}/samples"
        self._sample_url = urljoin(url, path)

        path = "catalogue/{session_id}/dataset"
        self._dataset_url = urljoin(url, path)

        path = "tracking/{session_id}/parcel"
        self._parcel_url = urljoin(url, path)

        path = "/session/{session_id}"
        self._session_info_url = urljoin(url, path)

        path = "session"
        self._authentication_url = urljoin(url, path)

        path = "samplemetadata/{session_id}/samples/acquisition"
        self._sample_metadata_url = urljoin(url, path)

        self._session_id = None

        if password:
            _ = self.login(password)

        if session_id:
            self._session_id = session_id

    def login(
        self, password: str, username: Optional[str] = None, plugin: str = "esrf"
    ) -> dict:
        credentials = {
            "plugin": plugin,
            "password": password,
        }
        if username is not None:
            credentials["username"] = username
        response = requests.post(self._authentication_url, json=credentials)
        response.raise_for_status()
        authentication_response = response.json()
        self._session_id = authentication_response["sessionId"]
        return authentication_response

    @property
    def session_id(self) -> str:
        """
        :raises RuntimeError: No session ID is available.
        """
        if self._session_id:
            return self._session_id

        raise RuntimeError("Login is required.")

    def get_investigations_by(
        self,
        filter: Optional[str] = None,
        instrument_name: Optional[str] = None,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
        ids: Optional[str] = None,
    ) -> List[dict]:
        """Returns a list of investigations matching the provided criteria.

        API Reference: https://icatplus.esrf.fr/api-docs/#/Catalogue/get_catalogue__sessionId__investigation

        :raises RuntimeError: No session ID is available.
        """
        params = {
            "filter": filter,
            "instrumentName": instrument_name,
            "ids": ids,
            "startDate": start_date.strftime("%Y-%m-%d") if start_date else None,
            "endDate": end_date.strftime("%Y-%m-%d") if end_date else None,
        }

        # Remove None values from params
        params = {k: v for k, v in params.items() if v is not None}

        url = self._investigation_url.format(session_id=self.session_id)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def get_parcels_by(self, investigation_id: str) -> List[dict]:
        """Returns the list of parcels associated to an investigation.

        :raises RuntimeError: No session ID is available.
        """
        params = {"investigationId": investigation_id}

        url = self._parcel_url.format(session_id=self.session_id)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def get_datasets_by(
        self, investigation_id: Optional[str] = None, dataset_ids: Optional[str] = None
    ) -> List[dict]:
        """Returns the list of dataset associated to an investigation of by ids.

        API Reference: https://icatplus.esrf.fr/api-docs/#/Catalogue/get_catalogue__sessionId__dataset
        """
        if investigation_id is None and dataset_ids is None:
            raise ValueError(
                "Either 'investigationId' or 'dataset_ids' must be provided."
            )

        params = {"investigationIds": investigation_id, "datasetIds": dataset_ids}
        # Remove None values from params
        params = {k: v for k, v in params.items() if v is not None}

        url = self._dataset_url.format(session_id=self.session_id)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def get_session_information(self) -> dict:
        """Fetches and returns session information from ICAT.

        API Reference: https://icatplus.esrf.fr/api-docs/#/Session/get_session__sessionId_

        :raises RuntimeError: No session ID is available.
        """
        url = self._session_info_url.format(session_id=self.session_id)
        response = requests.get(url)
        response.raise_for_status()
        return response.json()

    def get_samples_by(
        self,
        investigation_id: Optional[str] = None,
        sample_ids: Optional[str] = None,
        investigationId: Optional[str] = None,  # noqa N803
        sampleIds: Optional[str] = None,  # noqa N803
    ) -> List[dict]:
        """Returns a list of samples matching the provided criteria.

        API Reference: https://icatplus.esrf.fr/api-docs/#/Catalogue/get_catalogue__sessionId__samples
        """
        investigation_id = deprecated_argument(
            "investigation_id", investigation_id, "investigationId", investigationId
        )
        sample_ids = deprecated_argument(
            "sample_ids", investigation_id, "sampleIds", sampleIds
        )

        if investigation_id is None and sample_ids is None:
            raise ValueError(
                "Either 'investigationId' or 'sampleIds' must be provided."
            )

        params = {
            "investigationId": investigation_id,
            "sampleIds": sample_ids,
        }

        # Remove None values from params
        params = {k: v for k, v in params.items() if v is not None}

        url = self._sample_url.format(session_id=self.session_id)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def get_sample_metadata_by(
        self, proposal: str, beamline: str
    ) -> List[SampleMetadata]:
        params = {"proposal": proposal, "beamline": beamline}

        url = self._sample_metadata_url.format(session_id=self.session_id)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()



import warnings
from functools import wraps
from typing import Any


def deprecated_property(property_method):
    """Decorator to mark properties as deprecated with a warning."""

    @wraps(property_method)
    def deprecated_method(self):
        warnings.warn(
            f"Property '{property_method.__name__}' is deprecated and may be removed in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        return property_method(self)

    return property(deprecated_method)


def deprecated_argument(
    new_name: str, new_value: Any, deprecated_name: str, deprecated_value: Any
) -> Any:
    """Function to handle new and deprecated function argument."""
    if deprecated_value is not None:
        if new_value is not None:
            raise ValueError(
                f"Argument '{deprecated_name}' is replaced by '{new_name}'."
            )
        new_value = deprecated_value
        warnings.warn(
            f"Argument '{deprecated_name}' is deprecated and may be removed in future versions. Use '{new_name}' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
    return new_value


def deprecated_method(func):
    """Decorator to mark methods as deprecated with a warning."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"Method '{func.__name__}' is deprecated and may be removed in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        return func(*args, **kwargs)

    return wrapper


def warn_deprecated_module(
    message: str = "This module is deprecated and may be removed in future versions.",
):
    warnings.warn(message, ImportWarning, stacklevel=2)



import os


def basename(path: str) -> str:
    while path and not os.path.basename(path):
        path = os.path.dirname(path)
    return os.path.basename(path)


def dirname(path: str) -> str:
    while path and not os.path.basename(path):
        path = os.path.dirname(path)
    return os.path.dirname(path)


def split(path: str) -> str:
    while path and not os.path.basename(path):
        path = os.path.dirname(path)
    return path.split(os.sep)


def markdir(path: str) -> str:
    if path and os.path.basename(path):
        path = os.path.join(path, "")
    return path



import json
import socket
import time
from contextlib import contextmanager

import xmltodict

from ...concurrency import GEVENT_PATCHED
from ...concurrency import Queue
from ...concurrency import spawn
from .misc import eprint


def get_open_port():
    s = socket.socket()
    try:
        s.bind(("", 0))
        return s.getsockname()[1]
    finally:
        s.close()


def wait_tcp_online(host, port, timeout=5):
    """Wait for a TCP port with a timeout.

    Raises a `gevent.Timeout` if the port was not found.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    try:
        while True:
            try:
                sock.connect((host, port))
                break
            except ConnectionError:
                pass
    finally:
        sock.close()


@contextmanager
def tcp_message_server(data_parser=None, validate_all=True, timeout=5):
    """Start a TCP server and yield a queue of events.
    Data packages are separated by newline characters.
    Supported package encodings are UTF8 (default) and json.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    port = get_open_port()
    sock.bind(("localhost", port))
    sock.listen()

    # Listen to this socket
    messages = Queue()
    stop = False

    def listener():
        nonlocal stop

        buffer = b""
        conn = None
        try:
            while True:
                try:
                    conn, addr = sock.accept()
                    break
                except socket.timeout:
                    time.sleep(0.1)
            while True:
                try:
                    buffer += conn.recv(16384)
                except ConnectionResetError:
                    stop = True
                except socket.timeout:
                    pass
                if buffer:
                    out, sep, buffer = buffer.rpartition(b"\n")
                    if sep:
                        for bdata in out.split(b"\n"):
                            if data_parser == "json":
                                messages.put(json.loads(bdata))
                            elif b'xmlns:tns="http://www.esrf.fr/icat"' in bdata:
                                data = xmltodict.parse(
                                    bdata.decode(),
                                    process_namespaces=True,
                                    namespaces={"http://www.esrf.fr/icat": None},
                                )
                                messages.put(data)
                            else:
                                messages.put(bdata.decode())
                if stop:
                    return
                time.sleep(0.1)
        finally:
            if conn is not None:
                conn.close()

    glistener = spawn(listener)
    try:
        yield port, messages
    finally:
        messages.put(StopIteration)
        stop = True
        if validate_all:
            for msg in iter(messages.get, StopIteration):
                eprint(f"Unvalidated message: {msg}")
        if GEVENT_PATCHED:
            glistener.kill()
        else:
            glistener.join()
        sock.close()
